import os

import requests

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .. import interactive_console
from ..entities.artifact import ArtifactFactory, FileArtifact, LinkArtifact
from ..utils import Utils


class Artifacts:
    def __init__(self, feature_set, stub):
        self._feature_set = feature_set
        self._stub = stub
        self._feature_set_major_version = int(feature_set.version.split(".")[0])

    def store_file(self, file_path, title=None, description=None):
        store_response = self._request_backend_storage(file_path, title, description)
        interactive_console.log("Uploading file content")
        self._upload_content(file_path, store_response)
        retrieve_response = self._retrieve_artifact(store_response)
        interactive_console.log(f"File: {file_path} was successfully uploaded")
        return FileArtifact(retrieve_response.artifact, self._stub)

    def _request_backend_storage(self, file_path, title, description):
        file_name = os.path.basename(file_path)
        checksum = Utils.generate_md5_checksum(file_path)
        request = pb.StoreFileArtifactRequest(
            title=title or f"File {file_name}",
            description=description or "user uploaded file",
            filename=file_name,
            feature_set_id=self._feature_set.id,
            feature_set_major_version=self._feature_set_major_version,
            md5_checksum=checksum,
        )
        store_response = self._stub.StoreFileArtifact(request)
        return store_response

    def _upload_content(self, file_path, store_response):
        self._update_status(store_response.artifact_id, pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_IN_PROGRESS)
        with open(file_path, "rb") as file:
            data_response = requests.put(
                url=store_response.url,
                data=file,
                headers=store_response.headers,
            )
            if data_response.status_code not in range(200, 300):
                self._update_status(store_response.artifact_id, pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_FAILED)

                raise Exception(
                    f"File upload failed with status code {data_response.status_code} "
                    f"and message {data_response.text}"
                )
        self._update_status(store_response.artifact_id, pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_DONE)

    def _retrieve_artifact(self, store_response):
        retrieve_request = pb.RetrieveArtifactRequest(artifact_id=store_response.artifact_id)
        retrieve_response = self._stub.RetrieveArtifact(retrieve_request)
        return retrieve_response

    def _update_status(self, artifact_id, status):
        status_update_request = pb.UpdateUploadStatusRequest(artifact_id=artifact_id, upload_status=status)
        self._stub.UpdateUploadStatus(status_update_request)

    def store_link(self, url, title=None, description=None):
        upload_response = self._upload_link(url, title, description)
        retrieve_response = self._retrieve_artifact(upload_response)
        interactive_console.log(f"Link: {url} was successfully uploaded")
        return LinkArtifact(retrieve_response.artifact, self._stub)

    def _upload_link(self, url, title, description):
        request = pb.StoreLinkRequest(
            title=title or f"Link {url}",
            description=description or "user uploaded link",
            url=url,
            feature_set_id=self._feature_set.id,
            feature_set_major_version=self._feature_set_major_version,
        )
        response = self._stub.StoreLink(request)
        return response

    def list(self):
        """Return a generator which obtains artifacts associated with the feature set, lazily.

        Returns:
            Iterable[Artifact]: A generator iterator object consists of artifacts.

        Typical example:
            feature_set.artifacts.list()
        """
        request = pb.ListArtifactsRequest(
            feature_set_id=self._feature_set.id, feature_set_major_version=self._feature_set_major_version
        )
        while request:
            response = self._stub.ListArtifacts(request)
            if response.next_page_token:
                request.page_token = response.next_page_token
            else:
                request = None
            for artifact in response.artifacts:
                yield ArtifactFactory.create_acrtifact(artifact, self._stub)

    def __repr__(self):
        return "This class wraps together methods working with feature set artifacts"



from abc import ABC, abstractmethod

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from .utils import Utils


class StorageOptimization(ABC):
    @abstractmethod
    def _initialize(self, grpc: CoreServiceStub):
        raise NotImplementedError("Method `_initialize` needs to be implemented by the child class")

    @abstractmethod
    def _to_proto(self):
        raise NotImplementedError("Method `to_proto` needs to be implemented by the child class")

    @staticmethod
    def from_proto(proto: pb.StorageOptimization):
        if proto.HasField("compact"):
            return CompactOptimization()
        elif proto.HasField("z_order_by"):
            columns = proto.z_order_by.columns
            return ZOrderByOptimization(columns)
        else:
            return None

    def __repr__(self):
        return Utils.pretty_print_proto(self._to_proto())


class ZOrderByOptimization(StorageOptimization):
    def __init__(self, columns):
        self.columns = columns

    def _initialize(self, grpc: CoreServiceStub):
        pass

    def _to_proto(self):
        spec = pb.OptimizeStorageZOrderBySpec(columns=self.columns)
        proto = pb.StorageOptimization()
        proto.z_order_by.CopyFrom(spec)
        return proto


class CompactOptimization(StorageOptimization):
    def __init__(self):
        pass

    def _initialize(self, grpc: CoreServiceStub):
        pass

    def _to_proto(self):
        spec = pb.OptimizeStorageCompactSpec()
        proto = pb.StorageOptimization()
        proto.compact.CopyFrom(spec)
        return proto



from enum import Enum

from ai.h2o.featurestore.api.v1 import CoreService_pb2 as pb


class SearchOperator(Enum):
    SEARCH_OPERATOR_LIKE = 1
    SEARCH_OPERATOR_EQ = 2

    @classmethod
    def to_proto(cls, search_operator):
        return {
            cls.SEARCH_OPERATOR_LIKE: pb.AdvancedSearchOption.SearchOperator.SEARCH_OPERATOR_LIKE,
            cls.SEARCH_OPERATOR_EQ: pb.AdvancedSearchOption.SearchOperator.SEARCH_OPERATOR_EQ,
        }[search_operator]



import logging
import webbrowser

from google.protobuf.empty_pb2 import Empty


class Browser:
    _instance = None
    _web_url = None

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = object.__new__(cls)

        return cls._instance

    def __init__(self, stub):
        if self._web_url is None:
            response = stub.GetWebConfig(Empty())
            self._web_url = response.web_url

    def open_website(self, page=None):
        """This opens the Feature Store Web UI.

        This opens the returned URL in the browser (if it's not possible to open the browser, you will
        have to do this manually).

        For more details:
            https://docs.h2o.ai/featurestore/api/client_initialization
        """
        try:
            url = self._web_url + page if page else self._web_url
            webbrowser.get()
            webbrowser.open(url)
            logging.info(f"Opening browser to visit: {url}")
        except webbrowser.Error:
            logging.error(f"Browser is not supported. Please open " f"{url} manually.")



from ..utils import Utils
from .base_job import BaseJob


class OptimizeStorageJob(BaseJob):
    def _response_method(self, job_id):
        response = self._stub.GetOptimizeStorageJobOutput(job_id)
        return OptimizeStorageResponse(response)


class OptimizeStorageResponse:
    def __init__(self, response):
        self._response = response
        self._optimization_metrics = response.optimization_metrics

    @property
    def optimization_metrics(self):
        return self._optimization_metrics

    def __repr__(self):
        return Utils.pretty_print_proto(self._response)



import os
from enum import Enum

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from .. import interactive_console
from ..utils import Utils


class Artifact:
    def __init__(self, artifact, stub):
        self._artifact = artifact
        self._stub = stub

    @property
    def id(self):
        return self._artifact.artifact_id

    @property
    def title(self):
        return self._artifact.title

    @property
    def description(self):
        return self._artifact.description

    @property
    def artifact_type(self):
        return ArtifactType.from_proto(self._artifact.artifact_type)

    def delete(self):
        request = pb.DeleteArtifactRequest(artifact_id=self._artifact.artifact_id)
        self._stub.DeleteArtifact(request)
        interactive_console.log(f"Artifact '{self._artifact.title}' was deleted")

    def __repr__(self):
        return Utils.pretty_print_proto(self._artifact)


class FileArtifact(Artifact):
    def __init__(self, artifact, stub):
        super(FileArtifact, self).__init__(artifact, stub)

    @property
    def filename(self):
        return self._artifact.filename

    @property
    def upload_status(self):
        return ArtifactUploadStatus.from_proto(self._artifact.upload_status)

    def retrieve(self, filepath, overwrite_existing_file="n"):
        if not Utils.filepath_directory_exists(filepath):
            raise Exception(f"Can't save artifact as path does not exist: '{filepath}'")

        interactive_console.log("Requesting download link")
        request = pb.RetrieveArtifactRequest(artifact_id=self._artifact.artifact_id)
        response = self._stub.RetrieveArtifact(request)

        destination = os.path.join(filepath, response.artifact.filename) if os.path.isdir(filepath) else filepath
        if os.path.exists(destination) and overwrite_existing_file.lower() != "y":
            raise Exception(f"Artifact already exists '{destination}'")

        interactive_console.log("Downloading content")
        Utils.download_file(destination, response.artifact.url)
        interactive_console.log(f"Artifact '{self._artifact.title}' was saved into '{destination}'")


class LinkArtifact(Artifact):
    def __init__(self, artifact, stub):
        super(LinkArtifact, self).__init__(artifact, stub)

    @property
    def url(self):
        return self._artifact.url


class ArtifactFactory:
    @staticmethod
    def create_acrtifact(artifact, stub):
        if artifact.artifact_type == ArtifactType.ARTIFACT_TYPE_FILE.value:
            return FileArtifact(artifact, stub)
        elif artifact.artifact_type == ArtifactType.ARTIFACT_TYPE_LINK.value:
            return LinkArtifact(artifact, stub)
        else:
            raise Exception("Unknown artifact type encountered!")


class ArtifactType(Enum):
    ARTIFACT_TYPE_UNSPECIFIED = 0
    ARTIFACT_TYPE_LINK = 1
    ARTIFACT_TYPE_FILE = 2

    @classmethod
    def from_proto(cls, proto_artifacts_type):
        return {
            pb.ArtifactType.ARTIFACT_TYPE_UNSPECIFIED: cls.ARTIFACT_TYPE_UNSPECIFIED,
            pb.ArtifactType.ARTIFACT_TYPE_LINK: cls.ARTIFACT_TYPE_LINK,
            pb.ArtifactType.ARTIFACT_TYPE_FILE: cls.ARTIFACT_TYPE_FILE,
        }[proto_artifacts_type]


class ArtifactUploadStatus(Enum):
    ARTIFACT_UPLOAD_STATUS_NOT_APPLICABLE = 0
    ARTIFACT_UPLOAD_STATUS_IN_PROGRESS = 1
    ARTIFACT_UPLOAD_STATUS_DONE = 2
    ARTIFACT_UPLOAD_STATUS_FAILED = 3

    @classmethod
    def from_proto(cls, proto_upload_status):
        return {
            pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_NOT_APPLICABLE: cls.ARTIFACT_UPLOAD_STATUS_NOT_APPLICABLE,
            pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_IN_PROGRESS: cls.ARTIFACT_UPLOAD_STATUS_IN_PROGRESS,
            pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_DONE: cls.ARTIFACT_UPLOAD_STATUS_DONE,
            pb.ArtifactUploadStatus.ARTIFACT_UPLOAD_STATUS_FAILED: cls.ARTIFACT_UPLOAD_STATUS_FAILED,
        }[proto_upload_status]



import datetime
import json
import re
import time
from typing import Optional

from google.protobuf.timestamp_pb2 import Timestamp

import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb
import ai.h2o.featurestore.api.v1.OnlineApi_pb2 as OnlineApi
from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from ...rest_stub import RestStub
from ..job_info import JobInfo
from ..utils import Utils
from .ml_data_feature import MLDatasetFeature


class OnlineToken:
    def __init__(self, ml_dataset_id, token, signature, valid_to: datetime.datetime):
        self.ml_dataset_id = ml_dataset_id
        self.token = token
        self.signature = signature
        self.valid_to = valid_to

    def is_valid_for(self, ml_dataset_id):
        return (
            self.valid_to > (datetime.datetime.utcnow() - datetime.timedelta(minutes=1))
            and self.ml_dataset_id == ml_dataset_id
        )


class MLDataset:
    def __init__(self, stub: CoreServiceStub, rest_stub: RestStub, ml_dataset_proto):
        self._online_retrieval_token: Optional[OnlineToken] = None
        ml_dataset = pb.MLDataset()
        ml_dataset.CopyFrom(ml_dataset_proto)
        self._ml_dataset = ml_dataset
        self._stub = stub
        self._rest_stub = rest_stub

    @property
    def id(self):
        return self._ml_dataset.id

    @property
    def name(self):
        return self._ml_dataset.name

    @property
    def description(self):
        return self._ml_dataset.description

    @property
    def primary_keys(self):
        return self._ml_dataset.primary_keys

    @description.setter
    def description(self, value):
        update_request = pb.UpdateMLDatasetRequest(
            ml_dataset_id=self.id,
            description=value,
        )
        ml_dataset_proto = self._stub.UpdateMLDataset(update_request)
        ml_dataset = pb.MLDataset()
        ml_dataset.CopyFrom(ml_dataset_proto)
        self._ml_dataset = ml_dataset

    @property
    def start_date_time(self):
        timestamp: Timestamp = self._ml_dataset.start_date_time
        if timestamp:
            return datetime.datetime.fromtimestamp(timestamp.ToMilliseconds() / 1000, tz=datetime.timezone.utc)
        else:
            return None

    @property
    def end_date_time(self):
        timestamp: Timestamp = self._ml_dataset.end_date_time
        if timestamp:
            return datetime.datetime.fromtimestamp(timestamp.ToMilliseconds() / 1000, tz=datetime.timezone.utc)
        else:
            return None

    @property
    def features(self):
        return [MLDatasetFeature(feature) for feature in self._ml_dataset.features]

    def delete(self):
        """Deletes ML Dataset."""
        request = pb.DeleteMLDatasetRequest(ml_dataset_id=self.id)
        self._stub.DeleteMLDataset(request)

    def download(self, output_dir=None):
        """Download files to the specified directory location.

        Args:
            output_dir: (str) A directory location as string. Default is None.

        Returns:
            str: A directory path where the files are downloaded.

        Typical usage example:
            dir = ml_dataset.download()

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#downloading-the-files-from-feature-store-from-the-ml-dataset
        """
        request = pb.RetrieveMLDatasetAsAsLinksRequest(ml_dataset_id=self.id)
        job_id = self._stub.StartRetrieveMLDatasetAsLinksJob(request)

        info = JobInfo(self._stub, job_id)
        while not self._stub.GetJob(job_id).done:
            info.show_progress()
            time.sleep(2)
        info.show_progress()  # there is possibility that some progress was pushed before finishing job
        response = self._stub.GetRetrieveMLDatasetAsLinksJobOutput(job_id)

        return Utils.download_files(output_dir, response.download_links)

    def as_spark_frame(self, spark_session):
        """Return a spark data frame.

        This generates a data frame of the retrieved data using a spark session.

        Args:
            spark_session: (SparkSession) Represents a spark session.

        Returns:
            DataFrame: Represents a spark data frame.

            The frame is made up of columns as features and rows as records.
            Rows contain data within retrieve scope (filtered).

        Typical usage example:
            data_frame = ml_dataset.as_spark_frame(sparkSession)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#obtaining-data-as-a-spark-frame-from-the-ml-dataset
        """
        from ..commons.spark_utils import SparkUtils

        session_id = spark_session.conf.get("ai.h2o.featurestore.sessionId", "")
        request = pb.RetrieveMLDatasetAsSparkRequest(ml_dataset_id=self.id, session_id=session_id)

        response = self._stub.RetrieveMLDatasetAsSpark(request)
        spark_session.conf.set("ai.h2o.featurestore.sessionId", response.session_id)

        SparkUtils.configure_user_spark(spark_session)
        if response.offline_storage_format == pb.OfflineStorageFormat.OFFLINE_STORAGE_FORMAT_SNOWFLAKE:
            df = self.__load_data_from_snowflake(response, spark_session)
        else:
            df = self.__load_data_from_parquet(response, spark_session)

        return df

    def __load_data_from_parquet(self, response, spark_session):
        for k, v in response.options.items():
            spark_session.conf.set(k, v)
        df = spark_session.read.format("parquet").load(response.path)
        return df

    def __load_data_from_snowflake(self, response, spark_session):
        df = (
            spark_session.read.format("net.snowflake.spark.snowflake")
            .options(**response.options)
            .option("dbtable", response.path)
            .load()
        )
        for col_name in df.schema.names:
            df = df.withColumnRenamed(col_name, re.sub(r"(^_|_$)", "", col_name.replace('"', "")))

        return df

    def retrieve_online(self, *keys):
        """Retrieve data from the online Feature Store.

        Args:
            keys: (Any) A specific primary key value for which the entry is obtained.

        Returns:
            dict: A dictionary of specific instance (JSON row).

            For example:

            {'id': '01', 'department': 'Engineering', 'name': 'Test'}

        Typical example:
            ml_dataset.retrieve_online(1)

        For more details:
            https://docs.h2o.ai/feature-store/latest-stable/docs/api/feature_view_api.html#retrieving-data-from-online-feature-store
        """
        if self._online_retrieval_token is None or not self._online_retrieval_token.is_valid_for(self.id):
            request = OnlineApi.OnlineMLDataSetRetrieveTokenRequest(ml_dataset_id=self.id)

            response: OnlineApi.OnlineRetrieveTokenResponse = self._stub.GenerateMLDatasetOnlineRetrievalToken(request)
            self._online_retrieval_token = OnlineToken(
                ml_dataset_id=self.id,
                token=response.token,
                signature=response.signature,
                valid_to=Utils.timestamp_to_datetime(response.valid_to),
            )

        response = self._rest_stub.online_store.get(
            f"/v1/retrieve/ml-datasets/{self.id}",
            params={
                "token": self._online_retrieval_token.token,
                "signature": self._online_retrieval_token.signature,
                "key": map(lambda x: str(x), keys),
            },
        )
        json_row = response["row"]

        return json.loads(json_row)

    def __repr__(self):
        return Utils.pretty_print_proto(self._ml_dataset)

    def __str__(self):
        nl = "\n"
        return (
            f"Name                  : {self.name} \n"
            f"Description           : {self.description} \n"
            f"Feature view version  : {self._ml_dataset.feature_view_version} \n"
            f"State                 : {self._ml_dataset.state} \n"
            f"Primary keys          : {self.primary_keys} \n"
            f"Features              : \n{nl.join(self._custom_feature_fields())} \n"
        )

    def _custom_feature_fields(self):
        tmp_list = list()
        for feature in self.features:
            s = (
                "   {\n"
                f"     Name             : {feature.name} \n"
                f"     Data type        : {feature.data_type} \n"
                f"     Is primary key?  : {feature.is_primary_key} \n"
                "   }"
            )
            tmp_list.append(s)
        return tmp_list



import ai.h2o.featurestore.api.v1.CoreService_pb2 as pb

from ..search_field import SearchField
from ..search_operator import SearchOperator


class AdvancedSearchOption:
    def __init__(
        self,
        search_operator: SearchOperator,
        search_field: SearchField,
        search_value: str,
    ):
        self._search_operator = search_operator
        self._search_field = search_field
        self._search_value = search_value

    def _to_proto(self):
        return pb.AdvancedSearchOption(
            search_operator=SearchOperator.to_proto(self._search_operator),
            search_field=SearchField.to_proto(self._search_field),
            search_value=self._search_value,
        )



from ai.h2o.featurestore.api.v1.CoreService_pb2_grpc import CoreServiceStub

from .base_job import BaseJob
from .ml_dataset import MLDataset


class CreateMLDatasetJob(BaseJob):
    def __init__(self, stub: CoreServiceStub, rest_stub, job_id):
        super().__init__(stub, job_id)
        self._rest_stub = rest_stub

    def _response_method(self, job_id):
        return MLDataset(self._stub, self._rest_stub, self._stub.GetMLDatasetCreationJobOutput(job_id))



import configparser
import logging
import os


class AzureKeyCredentials:
    def __init__(self, account_name, account_key):
        self.account_name = account_name
        self.account_key = account_key


class AzureSasCredentials:
    def __init__(self, account_name, sas_token):
        self.account_name = account_name
        self.sas_token = sas_token


class AzurePrincipalCredentials:
    def __init__(self, account_name, client_id, tenant_id, secret):
        self.account_name = account_name
        self.client_id = client_id
        self.tenant_id = tenant_id
        self.secret = secret


class S3Credentials:
    def __init__(self, access_key, secret_key, region=None, endpoint=None, role_arn=None, session_token=None):
        if region is None:
            logging.warning("AWS region is not configured. Using 'us-east-1'.")
            self.region = "us-east-1"
        else:
            self.region = region
        self.access_token = access_key
        self.secret_token = secret_key
        self.session_token = session_token

        self.role_arn = role_arn
        self.endpoint = endpoint

    @classmethod
    def from_profile(cls, profile):
        if os.path.isfile(os.path.join(os.path.expanduser("~"), ".aws/credentials")):
            config = configparser.RawConfigParser()
            config.read(os.path.join(os.path.expanduser("~"), ".aws/credentials"))
            aws_key = config.get(profile, "aws_access_key_id")
            aws_secret = config.get(profile, "aws_secret_access_key")
            aws_session = config.get(profile, "aws_session_token", fallback=None)
            config.read(os.path.join(os.path.expanduser("~"), ".aws/config"))
            profile_entry = "default" if profile == "default" else f"profile {profile}"
            role_arn = config.get(profile_entry, "role_arn", fallback=None)
            endpoint = config.get(profile_entry, "endpoint_url", fallback=None)
            aws_region = config.get(profile_entry, "region", fallback=None)

            return S3Credentials(
                access_key=aws_key,
                secret_key=aws_secret,
                session_token=aws_session,
                region=aws_region,
                role_arn=role_arn,
                endpoint=endpoint,
            )
        else:
            return None


class SnowflakeCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class SnowflakeKeyPairCredentials:
    def __init__(self, user, private_key_file, passphrase=None):
        self.user = user
        self.private_key_file = private_key_file
        self.passphrase = passphrase


class TeradataCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class PostgresCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password


class MongoDbCredentials:
    def __init__(self, user, password):
        self.user = user
        self.password = password



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ai/h2o/featurestore/api/v1/OnlineApi.proto
# Protobuf Python Version: 4.25.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*ai/h2o/featurestore/api/v1/OnlineApi.proto\x12\x1a\x61i.h2o.featurestore.api.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"%\n\x16OnlineRetrieveResponse\x12\x0b\n\x03row\x18\x01 \x01(\t\"H\n\x16OnlineIngestionRequest\x12\r\n\x05token\x18\x03 \x01(\t\x12\x11\n\tsignature\x18\x04 \x01(\t\x12\x0c\n\x04rows\x18\x05 \x03(\t\")\n\x17OnlineIngestionResponse\x12\x0e\n\x06\x65rrors\x18\x01 \x03(\t\"R\n\x1bOnlineIngestionTokenRequest\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12\x1b\n\x13\x66\x65\x61ture_set_version\x18\x02 \x01(\t\"n\n\x1cOnlineIngestionTokenResponse\x12\r\n\x05token\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\t\x12,\n\x08valid_to\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"Q\n\x1aOnlineRetrieveTokenRequest\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12\x1b\n\x13\x66\x65\x61ture_set_version\x18\x02 \x01(\t\"<\n#OnlineMLDataSetRetrieveTokenRequest\x12\x15\n\rml_dataset_id\x18\x01 \x01(\t\"m\n\x1bOnlineRetrieveTokenResponse\x12\r\n\x05token\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\t\x12,\n\x08valid_to\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"w\n\x13OnlineFeatureSchema\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\t\x12?\n\x06nested\x18\x03 \x03(\x0b\x32/.ai.h2o.featurestore.api.v1.OnlineFeatureSchema\"\xf5\x02\n\x0bOnlineToken\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12,\n\x08valid_to\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x03 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x04 \x01(\x05\x12\x14\n\x0cprimary_keys\x18\x05 \x03(\t\x12\x1a\n\x12time_travel_column\x18\x06 \x01(\t\x12!\n\x19time_travel_column_format\x18\x07 \x01(\t\x12\x14\n\x0ctime_to_live\x18\x08 \x01(\x05\x12\x1d\n\x15time_to_live_interval\x18\t \x01(\t\x12!\n\x19online_to_offline_enabled\x18\n \x01(\x08\x12?\n\x06schema\x18\x0b \x03(\x0b\x32/.ai.h2o.featurestore.api.v1.OnlineFeatureSchema\"\x90\x02\n\x13OnlineRetrieveToken\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12,\n\x08valid_to\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x03 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x04 \x01(\x05\x12P\n\x0cprimary_keys\x18\x05 \x03(\x0b\x32:.ai.h2o.featurestore.api.v1.OnlineRetrieveToken.PrimaryKey\x1a-\n\nPrimaryKey\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\t\"\x8a\x0c\n\x1cOnlineMLDatasetRetrieveToken\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12,\n\x08valid_to\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x15\n\rml_dataset_id\x18\x03 \x01(\t\x12\x14\n\x0cprimary_keys\x18\x04 \x03(\t\x12M\n\x05query\x18\x05 \x01(\x0b\x32>.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query\x12p\n\x18\x66\x65\x61ture_set_primary_keys\x18\x06 \x03(\x0b\x32N.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.FeatureSetPrimaryKeys\x12\x15\n\ronline_schema\x18\x07 \x01(\t\x12 \n\x18online_pipeline_location\x18\x08 \x01(\t\x1a\x8f\x07\n\x05Query\x12\x61\n\x11selected_features\x18\x01 \x03(\x0b\x32\x46.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Feature\x12Q\n\x04\x66rom\x18\x02 \x01(\x0b\x32\x43.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.From\x12Q\n\x04join\x18\x03 \x03(\x0b\x32\x43.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Join\x1aP\n\x04\x46rom\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x02 \x01(\x05\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x1a*\n\x07\x46\x65\x61ture\x12\x0e\n\x06prefix\x18\x01 \x01(\t\x12\x0f\n\x07\x66\x65\x61ture\x18\x02 \x01(\t\x1a\xfe\x03\n\x04Join\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x02 \x01(\x05\x12\r\n\x05\x61lias\x18\x03 \x01(\t\x12R\n\x02on\x18\x04 \x03(\x0b\x32\x46.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Join.On\x12_\n\tjoin_type\x18\x05 \x01(\x0e\x32L.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Join.JoinType\x1a\xb1\x01\n\x02On\x12T\n\x04left\x18\x01 \x01(\x0b\x32\x46.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Feature\x12U\n\x05right\x18\x02 \x01(\x0b\x32\x46.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.Query.Feature\"C\n\x08JoinType\x12\x1b\n\x17ONLINE_QUERY_JOIN_INNER\x10\x00\x12\x1a\n\x16ONLINE_QUERY_JOIN_LEFT\x10\x01\x1a\xf1\x01\n\x15\x46\x65\x61tureSetPrimaryKeys\x12\x16\n\x0e\x66\x65\x61ture_set_id\x18\x01 \x01(\t\x12!\n\x19\x66\x65\x61ture_set_major_version\x18\x02 \x01(\x05\x12n\n\x0bprimary_key\x18\x03 \x03(\x0b\x32Y.ai.h2o.featurestore.api.v1.OnlineMLDatasetRetrieveToken.FeatureSetPrimaryKeys.PrimaryKey\x1a-\n\nPrimaryKey\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tdata_type\x18\x02 \x01(\tB\x10\x42\x0eOnlineApiProtob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ai.h2o.featurestore.api.v1.OnlineApi_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\016OnlineApiProto'
  _globals['_ONLINERETRIEVERESPONSE']._serialized_start=107
  _globals['_ONLINERETRIEVERESPONSE']._serialized_end=144
  _globals['_ONLINEINGESTIONREQUEST']._serialized_start=146
  _globals['_ONLINEINGESTIONREQUEST']._serialized_end=218
  _globals['_ONLINEINGESTIONRESPONSE']._serialized_start=220
  _globals['_ONLINEINGESTIONRESPONSE']._serialized_end=261
  _globals['_ONLINEINGESTIONTOKENREQUEST']._serialized_start=263
  _globals['_ONLINEINGESTIONTOKENREQUEST']._serialized_end=345
  _globals['_ONLINEINGESTIONTOKENRESPONSE']._serialized_start=347
  _globals['_ONLINEINGESTIONTOKENRESPONSE']._serialized_end=457
  _globals['_ONLINERETRIEVETOKENREQUEST']._serialized_start=459
  _globals['_ONLINERETRIEVETOKENREQUEST']._serialized_end=540
  _globals['_ONLINEMLDATASETRETRIEVETOKENREQUEST']._serialized_start=542
  _globals['_ONLINEMLDATASETRETRIEVETOKENREQUEST']._serialized_end=602
  _globals['_ONLINERETRIEVETOKENRESPONSE']._serialized_start=604
  _globals['_ONLINERETRIEVETOKENRESPONSE']._serialized_end=713
  _globals['_ONLINEFEATURESCHEMA']._serialized_start=715
  _globals['_ONLINEFEATURESCHEMA']._serialized_end=834
  _globals['_ONLINETOKEN']._serialized_start=837
  _globals['_ONLINETOKEN']._serialized_end=1210
  _globals['_ONLINERETRIEVETOKEN']._serialized_start=1213
  _globals['_ONLINERETRIEVETOKEN']._serialized_end=1485
  _globals['_ONLINERETRIEVETOKEN_PRIMARYKEY']._serialized_start=1440
  _globals['_ONLINERETRIEVETOKEN_PRIMARYKEY']._serialized_end=1485
  _globals['_ONLINEMLDATASETRETRIEVETOKEN']._serialized_start=1488
  _globals['_ONLINEMLDATASETRETRIEVETOKEN']._serialized_end=3034
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY']._serialized_start=1879
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY']._serialized_end=2790
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_FROM']._serialized_start=2153
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_FROM']._serialized_end=2233
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_FEATURE']._serialized_start=2235
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_FEATURE']._serialized_end=2277
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN']._serialized_start=2280
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN']._serialized_end=2790
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN_ON']._serialized_start=2544
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN_ON']._serialized_end=2721
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN_JOINTYPE']._serialized_start=2723
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_QUERY_JOIN_JOINTYPE']._serialized_end=2790
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_FEATURESETPRIMARYKEYS']._serialized_start=2793
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_FEATURESETPRIMARYKEYS']._serialized_end=3034
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_FEATURESETPRIMARYKEYS_PRIMARYKEY']._serialized_start=1440
  _globals['_ONLINEMLDATASETRETRIEVETOKEN_FEATURESETPRIMARYKEYS_PRIMARYKEY']._serialized_end=1485
# @@protoc_insertion_point(module_scope)



import abc
import collections
import sys
import time
import uuid
from random import randint
from typing import List, Optional

import grpc
from requests import models
from requests.auth import AuthBase

from .auth import AuthException


class _ClientCallDetails(
    collections.namedtuple("_ClientCallDetails", ("method", "timeout", "metadata", "credentials")),
    grpc.ClientCallDetails,
):
    pass


class AuthClientInterceptor(grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor):
    def __init__(self, client):
        self._client = client

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        if client_call_details.method in (
            "/ai.h2o.featurestore.api.v1.CoreService/Login",
            "/ai.h2o.featurestore.api.v1.CoreService/GetAccessToken",
            "/ai.h2o.featurestore.api.v1.CoreService/GetVersion",
            "/ai.h2o.featurestore.api.v1.CoreService/GetApiConfig",
            "/ai.h2o.featurestore.api.v1.CoreService/GetWebConfig",
        ):
            return continuation(client_call_details, request_or_iterator)
        else:
            metadata = []
            if client_call_details.metadata is not None:
                metadata = list(client_call_details.metadata)
            is_success, token_or_error = self._client.auth._obtain_token()
            if is_success:
                sys.tracebacklimit = None
                metadata.append(
                    (
                        "authorization",
                        "Bearer " + token_or_error,
                    )
                )
            else:
                sys.tracebacklimit = 0
                raise AuthException(token_or_error)
            metadata.append(("request-id", str(uuid.uuid4())))
            new_client_call_details = _ClientCallDetails(
                client_call_details.method,
                self._client._client_config.timeout,
                metadata,
                client_call_details.credentials,
            )
            return continuation(new_client_call_details, request_or_iterator)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator)


class RestAuth(AuthBase):
    def __init__(self, client):
        self._client = client

    def __call__(self, r: models.PreparedRequest) -> models.PreparedRequest:
        is_success, token_or_error = self._client.auth._obtain_token()
        if is_success:
            sys.tracebacklimit = None
            r.headers["authorization"] = "Bearer " + token_or_error
        else:
            sys.tracebacklimit = 0
            raise AuthException(token_or_error)

        return r


class SleepingPolicy(abc.ABC):
    @abc.abstractmethod
    def sleep(self, try_i: int):
        """How long to sleep in milliseconds.

        Args:
            try_i: (int) The number of retry (starting from zero)
        """
        assert try_i >= 0


class ExponentialBackoff(SleepingPolicy):
    def __init__(self, *, init_backoff_ms: int, max_backoff_ms: int, multiplier: int):
        self.init_backoff = randint(0, init_backoff_ms)
        self.max_backoff = max_backoff_ms
        self.multiplier = multiplier

    def sleep(self, try_i: int):
        sleep_range = min(self.init_backoff * self.multiplier**try_i, self.max_backoff)
        sleep_ms = randint(0, sleep_range)
        time.sleep(sleep_ms / 1000)


class RetryOnRpcErrorClientInterceptor(grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor):
    def __init__(
        self,
        *,
        max_attempts: int,
        sleeping_policy: SleepingPolicy,
        status_for_retry: Optional[List[grpc.StatusCode]] = None,
    ):
        self.max_attempts = max_attempts
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):

        for try_i in range(self.max_attempts):
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):

                # Return if it was last attempt
                if try_i == (self.max_attempts - 1):
                    return response

                # If status code is not in retryable status codes
                if self.status_for_retry and response.code() not in self.status_for_retry:
                    return response

                self.sleeping_policy.sleep(try_i)
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept_call(continuation, client_call_details, request_iterator)



import uuid

import requests
from google.protobuf.empty_pb2 import Empty

from .core.interceptors import RestAuth


class RestExecutor:
    def __init__(self, client, base_url, root_certificates: str):
        self._client = client
        self._rest_auth = RestAuth(client)
        self._base_url = base_url
        self._root_certificates = root_certificates

    def post(self, url: str, body: str, params=None):
        response = requests.post(
            self._base_url + url,
            data=body,
            auth=self._rest_auth,
            params=params,
            verify=self._root_certificates if self._root_certificates else True,
            headers={
                "user-agent": f"feature-store-py-cli/{self._client._get_client_version()}",
                "content-type": "application/json",
                "request-id": str(uuid.uuid4()),
            },
        )

        response.raise_for_status()
        return response.json()

    def get(self, url: str, params=None):
        response = requests.get(
            self._base_url + url,
            auth=self._rest_auth,
            params=params,
            verify=self._root_certificates if self._root_certificates else True,
            headers={
                "user-agent": f"feature-store-py-cli/{self._client._get_client_version()}",
                "content-type": "application/json",
                "request-id": str(uuid.uuid4()),
            },
        )

        response.raise_for_status()
        return response.json()


class RestStub:
    def __init__(self, client, root_certificates: str):
        response = client._stub.GetApiConfig(Empty())
        self.online_store = RestExecutor(client, response.public_online_rest_api_url, root_certificates)
